"""
PYVED game engine by moonb3ndr [thomas.iw@kata.games] et al.

Visit https://kata.games to learn more!
"""


import pyved_engine as _engine
# _engine.bootstrap_e()

from pyved_engine import component, entity, EntityManager, System, SystemManager
from pyved_engine.core.events import EvManager, EvListener
from pyved_engine import game_events_enum, get_ev_manager, init, quit, get_surface, flip


def pygame_proxy():
    _engine.bootstrap_e()
    return _engine.pygame


# lists what should be interpreted by import *
# this overrides the default of hiding everything that begins with an underscore
__all__ = [
    component, entity, EntityManager, System, SystemManager,
    EvManager, EvListener,
    game_events_enum, get_ev_manager, init, quit, get_surface, flip,
    pygame_proxy
]



import pygame
from dataclasses import dataclass


@dataclass
class Person:
    name: str
    age: int
    height: float
    email: str


mec = Person('thomas', 99, 1.85, 'thomas.iw@kata.games')
print('hello world')
gameover = False
surf = pygame.display.set_mode((640, 480))

while not gameover:
    for ev in pygame.event.get():
        if ev.type == pygame.QUIT:
            gameover = True
    surf.fill('pink')
    pygame.display.flip()

print('bye', mec.name, 'see you when you turn', mec.age+1)



from typing import Union, Dict, Tuple
import pygame_gui
from pygame_gui.core.interfaces import IContainerLikeInterface, IUIManagerInterface
from pygame_gui.core import UIElement, UIContainer
from pygame_gui.core.drawable_shapes import RectDrawableShape
from pygame_gui.elements.ui_button import UIButton
from pygame_gui.elements.ui_selection_list import UISelectionList
import pygame


menu_data = {'#file_menu': {'display_name': 'File',
                            'items':
                                {
                                    '#new': {'display_name': 'New...'},
                                    '#open': {'display_name': 'Open...'},
                                    '#save': {'display_name': 'Save'},
                                    '#save_as': {'display_name': 'Save As...'}
                                }
                            },
             '#edit_menu': {'display_name': 'Edit',
                            'items':
                                {
                                    '#undo': {'display_name': 'Undo'},
                                    '#redo': {'display_name': 'Redo'}
                                }
                            },
             '#view_menu': {'display_name': 'View',
                            'items':
                                {
                                    '#info': {'display_name': 'Image info'}
                                }
                            },
             '#help_menu': {'display_name': 'Help',
                            'items':
                                {
                                    '#about': {'display_name': 'About'}
                                }
                            }
             }


class UIMenuBar(UIElement):
    def __init__(self,
                 relative_rect: pygame.Rect,
                 menu_item_data,
                 manager: IUIManagerInterface,
                 container: Union[IContainerLikeInterface, None] = None,
                 parent: Union[UIElement, None] = None,
                 object_id: Union[str, None] = None,
                 anchors: Union[Dict[str, str], None] = None):

        super().__init__(relative_rect=relative_rect,
                         manager=manager,
                         container=container,
                         starting_height=1,
                         layer_thickness=1,
                         anchors=anchors)

        self._create_valid_ids(container=container,
                               parent_element=parent,
                               object_id=object_id,
                               element_id='menu_bar')

        self.menu_item_data = menu_item_data

        self.background_colour = None
        self.border_colour = None
        self.shape_type = 'rectangle'

        self.open_menu = None
        self._selected_menu_bar_button = None

        self._close_open_menu = False

        self.rebuild_from_changed_theme_data()

        container_rect = pygame.Rect(self.relative_rect.left +
                                     (self.shadow_width + self.border_width),
                                     self.relative_rect.top +
                                     (self.shadow_width + self.border_width),
                                     self.relative_rect.width -
                                     2 * (self.shadow_width + self.border_width),
                                     self.relative_rect.height -
                                     2 * (self.shadow_width + self.border_width))

        self.menu_bar_container = UIContainer(container_rect, manager,
                                              starting_height=1,
                                              parent_element=self,
                                              object_id='#menu_bar_container')

        top_level_button_x = 0
        for menu_item_key, menu_item_data in self.menu_item_data.items():
            # create top level menu buttons
            default_font = self.ui_manager.get_theme().get_font_dictionary().get_default_font()
            item_text_rect = default_font.render_premul(menu_item_data['display_name'], (0, 0, 0))
            UIButton(pygame.Rect(top_level_button_x,
                                 0,
                                 item_text_rect.get_width() + 20,
                                 self.menu_bar_container.rect.height),
                     text=menu_item_data['display_name'],
                     manager=self.ui_manager,
                     container=self.menu_bar_container,
                     object_id=menu_item_key,
                     parent_element=self)
            top_level_button_x += item_text_rect.get_width() + 20

    def unfocus(self):
        if self.open_menu is not None:
            self.open_menu.kill()
            self.open_menu = None
        if self._selected_menu_bar_button is not None:
            self._selected_menu_bar_button.unselect()
            self._selected_menu_bar_button = None

    def update(self, time_delta: float):
        """
        A method called every update cycle of our application. Designed to be overridden by derived
        classes but also has a little functionality to make sure the menu's layer 'thickness' is
        accurate and to handle window resizing.

        :param time_delta: time passed in seconds between one call to this method and the next.

        """
        super().update(time_delta)
        if self.menu_bar_container.layer_thickness != self.layer_thickness:
            self.layer_thickness = self.menu_bar_container.layer_thickness

    def process_event(self, event: pygame.event.Event) -> bool:
        """
        Can be overridden, also handle resizing windows. Gives UI Windows access to pygame events.
        Currently just blocks mouse click down events from passing through the menu.

        :param event: The event to process.

        :return: Should return True if this element consumes this event.

        """
        consumed_event = False
        if (self is not None and
                event.type == pygame.MOUSEBUTTONDOWN and
                event.button in [pygame.BUTTON_LEFT,
                                 pygame.BUTTON_RIGHT,
                                 pygame.BUTTON_MIDDLE]):
            scaled_mouse_pos = (int(event.pos[0] * self.ui_manager.mouse_pos_scale_factor[0]),
                                int(event.pos[1] * self.ui_manager.mouse_pos_scale_factor[1]))

            if self.hover_point(scaled_mouse_pos[0], scaled_mouse_pos[1]):
                consumed_event = True

        if (event.type == pygame_gui.UI_BUTTON_START_PRESS and
                event.ui_element in self.menu_bar_container.elements):
            if self._selected_menu_bar_button is not None:
                self._selected_menu_bar_button.unselect()
            self._selected_menu_bar_button = event.ui_element
            self._selected_menu_bar_button.select()
            self._open_top_level_menu(event)

        if (event.type == pygame_gui.UI_BUTTON_ON_HOVERED and
                self.open_menu is not None and
                event.ui_element in self.menu_bar_container.elements):
            if self._selected_menu_bar_button is not None:
                self._selected_menu_bar_button.unselect()
            self._selected_menu_bar_button = event.ui_element
            self._selected_menu_bar_button.select()
            self._open_top_level_menu(event)

        return consumed_event

    def _open_top_level_menu(self, event):
        # kill any open menus
        if self.open_menu is not None:
            self.open_menu.kill()
        # open this menu
        menu_key = event.ui_object_id.split('.')[-1]
        menu_size = ((len(self.menu_item_data[menu_key]['items']) * 20) +
                     (2 * (0 + 1)))
        item_data = [
            (item_data['display_name'], item_key)
            for item_key, item_data in self.menu_item_data[menu_key][
                'items'
            ].items()
        ]
        menu_rect = pygame.Rect(0, 0, 200, menu_size)
        menu_rect.topleft = event.ui_element.rect.bottomleft
        top_ui_layer = self.ui_manager.get_sprite_group().get_top_layer()
        self.open_menu = UISelectionList(relative_rect=menu_rect,
                                         item_list=item_data,
                                         manager=self.ui_manager,
                                         starting_height=top_ui_layer,
                                         parent_element=self,
                                         object_id=menu_key + '_items')
        self.ui_manager.set_focus_set(self)

    def kill(self):
        """
        Overrides the basic kill() method of a pygame sprite so that we also kill all the UI
        elements in this menu.

        """
        self.menu_bar_container.kill()
        super().kill()

    def set_dimensions(self, dimensions: Union[pygame.math.Vector2,
                                               Tuple[int, int],
                                               Tuple[float, float]]):
        """
        Set the size of this menu and then re-sizes and shifts the contents of the menu container
        to fit the new size.

        :param dimensions: The new dimensions to set.

        """
        # Don't use a basic gate on this set dimensions method because the container may be a
        # different size to the window
        super().set_dimensions(dimensions)

        new_container_dimensions = (self.relative_rect.width -
                                    2 * (self.shadow_width + self.border_width),
                                    self.relative_rect.height -
                                    2 * (self.shadow_width + self.border_width))
        if new_container_dimensions != self.menu_bar_container.relative_rect.size:
            container_rel_pos = (self.relative_rect.x + (self.shadow_width + self.border_width),
                                 self.relative_rect.y + (self.shadow_width + self.border_width))
            self.menu_bar_container.set_dimensions(new_container_dimensions)
            self.menu_bar_container.set_relative_position(container_rel_pos)

    def set_relative_position(self, position: Union[pygame.math.Vector2,
                                                    Tuple[int, int],
                                                    Tuple[float, float]]):
        """
        Method to directly set the relative rect position of an element.

        :param position: The new position to set.

        """
        super().set_relative_position(position)
        container_rel_pos = (self.relative_rect.x + (self.shadow_width + self.border_width),
                             self.relative_rect.y + (self.shadow_width + self.border_width))
        self.menu_bar_container.set_relative_position(container_rel_pos)

    def set_position(self, position: Union[pygame.math.Vector2,
                                           Tuple[int, int],
                                           Tuple[float, float]]):
        """
        Method to directly set the absolute screen rect position of an element.

        :param position: The new position to set.

        """
        super().set_position(position)
        container_rel_pos = (self.relative_rect.x + (self.shadow_width + self.border_width),
                             self.relative_rect.y + (self.shadow_width + self.border_width))
        self.menu_bar_container.set_relative_position(container_rel_pos)

    def rebuild_from_changed_theme_data(self):
        """
        Checks if any theming parameters have changed, and if so triggers a full rebuild of the
        button's drawable shape.
        """
        has_any_changed = False

        background_colour = self.ui_theme.get_colour_or_gradient('normal_bg',
                                                                 self.combined_element_ids)
        if background_colour != self.background_colour:
            self.background_colour = background_colour
            has_any_changed = True

        border_colour = self.ui_theme.get_colour_or_gradient('normal_border',
                                                             self.combined_element_ids)
        if border_colour != self.border_colour:
            self.border_colour = border_colour
            has_any_changed = True

        # misc
        shape_type_string = self.ui_theme.get_misc_data('shape', self.combined_element_ids)
        if (shape_type_string is not None and shape_type_string in ['rectangle'] and
                shape_type_string != self.shape_type):
            self.shape_type = shape_type_string
            has_any_changed = True

        if self._check_shape_theming_changed(defaults={'border_width': 1,
                                                       'shadow_width': 0,
                                                       'shape_corner_radius': 0}):
            has_any_changed = True

        if has_any_changed:
            self.rebuild()

    def rebuild(self):
        """
        A complete rebuild of the drawable shape used by this button.

        """
        theming_parameters = {'normal_bg': self.background_colour,
                              'normal_border': self.border_colour,
                              'border_width': self.border_width,
                              'shadow_width': self.shadow_width}

        if self.shape_type == 'rectangle':
            self.drawable_shape = RectDrawableShape(self.rect, theming_parameters,
                                                    ['normal'], self.ui_manager)

        self.on_fresh_drawable_shape_ready()



import pygame
import pygame_gui

# from pygame_gui.ui_manager import UIManager
# from pygame_gui.elements.ui_window import UIWindow
from pygame_gui.elements.ui_image import UIImage
from pygame_gui import UIManager, UI_TEXT_ENTRY_CHANGED
from pygame_gui.elements import UIWindow, UITextEntryBox, UITextBox
from .menu_bar_gui import UIMenuBar, menu_data

from . import menu_bar_ev_handler

from .pong.pong import PongGame
import os  # so i can open file to read src code!
import pathlib

PONG_WINDOW_SELECTED = pygame.event.custom_type()


APP_W, APP_H = 1380, 800
KTG_W, KTG_H = 960, 720
# ktg sont les dim. d'un jeu
POS_PREVIEW = (200, 0)

target_dir = None
target_file = 'ERR_NO_FILE'


class GamePreview(UIWindow):
    def __init__(self, gamename, position, ui_manager):
        super().__init__(pygame.Rect(position, (KTG_W+32, KTG_H+59)), ui_manager,
                         window_display_title=gamename,
                         object_id='#pong_window',
                         draggable=False)

        game_surface_size = self.get_container().get_size()
        print('surf pour affichage jeu:', game_surface_size)

        self.game_surface_element = UIImage(pygame.Rect((0, 0),
                                                        game_surface_size),
                                            pygame.Surface(game_surface_size).convert(),
                                            manager=ui_manager,
                                            container=self,
                                            parent_element=self)

        self.pong_game = PongGame()
        self.is_active = False

    def process_event(self, event):
        handled = super().process_event(event)
        if (event.type == pygame_gui.UI_BUTTON_PRESSED and
                event.ui_object_id == "#pong_window.#title_bar" and
                event.ui_element == self.title_bar):
            handled = True
            event_data = {'ui_element': self,
                          'ui_object_id': self.most_specific_combined_id}
            window_selected_event = pygame.event.Event(PONG_WINDOW_SELECTED,
                                                       event_data)
            pygame.event.post(window_selected_event)
        if self.is_active:
            handled = self.pong_game.process_event(event)
        return handled

    def update(self, time_delta):
        if self.alive() and self.is_active:
            self.pong_game.update(time_delta)

        super().update(time_delta)

        self.pong_game.draw(self.game_surface_element.image)


def gen_edition_title(file, project_name):
    return file+f"({project_name}) source-code"


class MiniGamesApp:
    def __init__(self):
        pygame.init()
        pygame.display.set_caption('Pyved: game creation toolbox')
        self.root_window_surface = pygame.display.set_mode((APP_W, APP_H))

        self.background_surface = pygame.Surface((APP_W, APP_H)).convert()
        self.background_surface.fill(pygame.Color('#505050'))
        self.ui_manager = UIManager((APP_W, APP_H), os.path.join(os.path.dirname(__file__),'gui-rel-data/ui_theme.json'))
        self.clock = pygame.time.Clock()
        self.is_running = True

        self.pong_window_1 = GamePreview(
            'Pong demo',
            POS_PREVIEW,
            self.ui_manager
        )
        # self.pong_window_2 = PongWindow((50, 50), self.ui_manager)
        self.pong_window_1.is_active = True

        # --------- text edition related stuff -------------
        if target_dir is not None:
            path = pathlib.PurePath(target_dir)
            pure_dir = path.name
        else:
            pure_dir = '??'
        self.notepad_window = UIWindow(pygame.Rect(50, 20, 380, 400), window_display_title=gen_edition_title(target_file, pure_dir))
        output_window = UIWindow(pygame.Rect(440, 20, 380, 400), window_display_title="Score")

        self.text_entry_box = UITextEntryBox(  # swap to editable text box
                relative_rect=pygame.Rect((0, 0), self.notepad_window.get_container().get_size()),
                initial_text="",
                container=self.notepad_window)

        self.text_output_box = UITextBox(
                relative_rect=pygame.Rect((0, 0), output_window.get_container().get_size()),
                html_text="",
                container=output_window)
        if target_dir is not None:
            thefile = os.path.join(target_dir, target_file)
            with open(thefile, 'r') as fptr:
                self.text_entry_box.set_text(fptr.read())

        # -------------- menu bar related stuff ------------------
        self.menu_bar = UIMenuBar(relative_rect=pygame.Rect(0, 0, 1280, 25),
                                  menu_item_data=menu_data,
                                  manager=self.ui_manager)
        self.menu_bar_event_handler = menu_bar_ev_handler.MenuBarEventHandler(self.root_window_surface, self.ui_manager)

    def run(self):
        while self.is_running:
            time_delta = self.clock.tick(60)/1000.0

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.is_running = False

                self.ui_manager.process_events(event)

                # use the menu bar also!
                self.menu_bar_event_handler.process_event(event)

                if event.type == PONG_WINDOW_SELECTED:
                    event.ui_element.is_active = True
##                    if event.ui_element == self.pong_window_1:
##                        self.pong_window_2.is_active = False
##                    elif event.ui_element == self.pong_window_2:
##                        self.pong_window_1.is_active = False
                elif event.type == UI_TEXT_ENTRY_CHANGED and event.ui_element == self.text_entry_box:
                    self.text_output_box.set_text(event.text)

                # --- open pyved.json --- {
                elif (event.type == pygame_gui.UI_FILE_DIALOG_PATH_PICKED and
                        event.ui_object_id == '#open_file_dialog'):
                    path = pathlib.Path(event.text)
                    p = pathlib.PurePath(event.text)
                    filename = p.name
                    print(path.parent)
                    p2 = pathlib.PurePath(path.parent)
                    adhoc_dir = p2.name

                    self.menu_bar_event_handler.last_used_file_path = path.parent
                    try:
                        self.notepad_window.set_display_title(gen_edition_title(filename, adhoc_dir))
                        with open(str(path), 'r') as fptr:
                            self.text_entry_box.set_text(fptr.read())
##                        loaded_image = pygame.image.load(str(path)).convert_alpha()
##
##                        canvas_window_rect = pygame.Rect(200, 25,
##                                                         min(loaded_image.get_width() + 52,
##                                                             self.window_surface.get_width() - 200),
##                                                         min(loaded_image.get_height() + 82,
##                                                             self.window_surface.get_height() - 25))
##
##                        window = CanvasWindow(rect=canvas_window_rect,
##                                              manager=self.ui_manager,
##                                              image_file_name=path.name,
##                                              image=loaded_image)
##
##                        window.canvas_ui.set_active_tool(self.tool_bar_window.get_active_tool())
##                        window.canvas_ui.set_save_file_path(path)
                    except pygame.error:
                        message_rect = pygame.Rect(0, 0, 250, 160)
                        message_rect.center = self.window_surface.get_rect().center
                        message_window = UIMessageWindow(rect=message_rect,
                                                         html_message='Unable to load image.',
                                                         manager=self.ui_manager,
                                                         window_title='Loading error')
                        message_window.set_blocking(True)
                    # fin open pyved.json

            self.ui_manager.update(time_delta)

            self.root_window_surface.blit(self.background_surface, (0, 0))
            self.ui_manager.draw_ui(self.root_window_surface)

            pygame.display.update()


def run_editor():
    app = MiniGamesApp()
    app.run()



import argparse
import os
from pathlib import Path
from . import utils


parser = argparse.ArgumentParser(
    prog='pyved',
    description='runs PYVED and opens a new/existing project',
    epilog="pyved (=PY Visual Editor) is a new toolbox for easy game development. For more information, please visit "
           "https://github.com/wktab/gamedev-pyved/"
)
# configuration --
# besoin d'utiliser des sub parsers car on a plusieurs cmd
subparser = parser.add_subparsers(dest='command', required=True)
new = subparser.add_parser('new')
new.add_argument('project', type=str)  # , required=True)
register = subparser.add_parser('open')
register.add_argument('filepath', type=str)  # , required=True)
cmd = subparser.add_parser('version')
# -- fin configuration


# definition implem {{
def create_new_project(pname, tool_ver):
    target_dir = Path(pname)
    if target_dir.exists():
        print(f"ERROR: folder {target_dir} already exists")
    else:
        os.mkdir(target_dir)
        print(f" +Initializing a new project named \"{target_dir}\"...")
        DB_QT = '"'
        with open(os.path.join(target_dir, 'pyved.json'), 'w') as fout:
            fout.write("{\n")
            fout.write(' ' * 4 + '"tag":' + DB_QT + 'pyved' + tool_ver + DB_QT + ",\n")
            fout.write(' ' * 4 + '"project":' + DB_QT + pname + DB_QT + ",\n")
            fout.write("}")

        # main source-code file
        with open(os.path.join(target_dir, 'main.py'), 'w') as fout:
            fout.write("import katagames_engine as kengi")
            fout.write("\n")
            fout.write("kengi.init()")


def open_project(desc_ptr, dirname):
    print('*' * 32)
    print(' Welcome to PYVED')
    print('*' * 32)

    print()
    print(desc_ptr.read())
    print('---')

    print('Listing files:')

    t_dir = Path(dirname)
    has_main = False
    for entry in t_dir.iterdir():
        print(entry.name)
        if 'main.py' == entry.name:
            has_main = True
    if not has_main:
        raise Exception('invalid project format! main.py not found in the project folder')

    # ouverture projet avec format valide ds l'editor
    from . import editor
    editor.target_dir = dirname
    editor.target_file = 'main.py'
    editor.run_editor()


# }} definition implem


# ------
#  main chunk of code
# -----
def main():
    # tool_version = pyved.__version__
    tool_ver = utils.get_version()

    args = parser.parse_args()
    if args.command == 'new':
        create_new_project(args.project, tool_ver)

    elif args.command == 'open':
        filepath = os.path.abspath(args.filepath)
        dirname = os.path.dirname(filepath)
        if not os.path.isfile(os.path.join(dirname, filepath)):
            print(f"ERROR: cannot read the file {filepath}, please provide a valid pyved.json FILE")
        else:
            fptr = open(filepath, 'r')
            open_project(fptr, dirname)
            fptr.close()

    elif args.command == 'version':
        print(f"PYVED - version {tool_ver}")
    else:
        print(f"ERROR: command {args.command} is not a valid command!")


if __name__ == '__main__':
    main()



"""
ECS - Entity Component system
heavily inspired by other projects authored by:
 - Vladimir Kaukin [KaukinVK@ya.ru]
 - Rik Cross [rik@raspberrypi.org]
"""
# from typing import Iterable, Iterator, Any
import dataclasses
import collections
import functools


# any entity class must be decorated with this function
entity = functools.partial(dataclasses.dataclass, slots=True)

# any component class must be decorated with this function
component = dataclasses.dataclass


class EntityManager:
    """Entity manager"""

    def __init__(self):
        self._entity_map = {}  # Person: [ent1, ent2]
        self._entity_components_map = {}  # Person: (MoveCom, DamageCom, NameCom)
        self._set_cache_map = {}  # (MoveCom, DamageCom, NameCom): {MoveCom, DamageCom, NameCom}
        self._delete_entity_buffer = collections.deque()  # deque([Person1, Person2])

    def add(self, *entity_value_list):
        """Add entities to world"""
        for entity_value in entity_value_list:
            assert getattr(entity_value, '__dict__', None) in (None, {}), 'Data class with inefficient memory usage'
            entity_value_class = entity_value.__class__
            self._entity_map.setdefault(entity_value_class, []).append(entity_value)
            if entity_value_class not in self._entity_components_map:
                self._entity_components_map[entity_value_class] = tuple(sorted(
                    (i for i in entity_value_class.__mro__ if i is not object),
                    key=lambda x: x.__class__.__name__
                ))

    def delete(self, *entity_value_list):
        """Delete entities from world"""
        for entity_value in entity_value_list:
            self._entity_map[entity_value.__class__].remove(entity_value)

    def delete_buffer_add(self, *entity_value_list):
        """Save entities into delete buffer for delete them from world later"""
        for entity_value in entity_value_list:
            self._delete_entity_buffer.append(entity_value)

    def delete_buffer_purge(self):
        """Delete all entities from delete buffer"""
        for delete_entity in self._delete_entity_buffer:
            self.delete(delete_entity)
        self._delete_entity_buffer.clear()

    def init(self, *entity_list):
        """
        Let entity manager to "know" about entities before work
        If manager do not know about entity, it will raise KeyError on access to it.
        event: SomeEvent = next(self.entities.get_by_class(SomeEvent), None)
        """
        for ent in entity_list:
            self.add(ent)
            self.delete(ent)

    def get_by_class(self, *entity_class_val_list: type):
        """
        Get all entities by specified entity class in specified order
        raise KeyError for uninitialized (never added) entities
        type returned is
        -> Iterator[Any]
        """
        for entity_class_val in entity_class_val_list:
            yield from self._entity_map[entity_class_val]

    def get_with_component(self, *component_class_val_list: type):
        """
        Get all entities that contains all specified component classes
        Sometimes it will be useful to warm up the cache
        raise KeyError for uninitialized (never added) entities
        type returned is
        -> Iterator[Any]
        """
        for entity_class, entity_component_list in self._entity_components_map.items():
            entity_component_set = \
                self._set_cache_map.setdefault(entity_component_list, set(entity_component_list))
            component_class_val_set = \
                self._set_cache_map.setdefault(component_class_val_list, set(component_class_val_list))
            if component_class_val_set.issubset(entity_component_set):
                yield from self._entity_map[entity_class]


class System:
    """
    Abstract base class for system
    All systems must be derived from this class
    System should have data for work: implement __init__ method
    """

    def initialize(self):
        """
        Preparing system to work before starting SystemManager.systems_update loop
        Runs by SystemManager.start_systems
        """

    def proc(self):
        """
        Run main system logic
        Runs by SystemManager.update_systems
        """

    def cleanup(self):
        """
        Clean system resources after stop update loop
        Runs by SystemManager.stop_systems
        """


class SystemManager:
    """
    System manager
    """

    def __init__(self, system_list):
        """
        :param system_list: an ordered sequence with systems, expected type is Iterable[System]
        such that
        for each pair of elements elt_i, elt_j we have classname(elt_i) != classname(elt_j)
        """
        self._system_list = tuple(system_list)

        self._name_to_sys = dict()
        for e in self._system_list:
            cls_name = e.__class__.__name__
            if cls_name in self._name_to_sys:
                raise Exception('ERR: Duplicate name for system detected!', cls_name)
            self._name_to_sys[cls_name] = e

        self._system_with_start_list = tuple(e for e in self._system_list if hasattr(e, 'initialize'))
        self._system_with_update_list = tuple(e for e in self._system_list if hasattr(e, 'proc'))
        self._system_with_stop_list = tuple(e for e in self._system_list if hasattr(e, 'cleanup'))

    # get by system name
    def __getitem__(self, item):
        return self._name_to_sys[item]

    def init_all(self):
        """Start all systems"""
        for system in self._system_with_start_list:
            system.initialize()

    def proc_all(self):
        """Update all systems"""
        for system in self._system_with_update_list:
            system.proc()

    def cleanup_all(self):
        """Stop all systems"""
        for system in self._system_with_stop_list:
            system.cleanup()



import sys
from pathlib import Path
import os
from setuptools import setup


def get_version():
    with open(os.path.join('src','pyved','VERSION')) as version_file:
        return version_file.read().strip()


sys.path.append('src')
# import pyved

# - fetch data from requirements.txt
with open('requirements.txt') as f:
    required = f.read().splitlines()

# read the contents of the README file


this_directory = Path(__file__).parent
long_desc = (this_directory / "README.md").read_text()

setup(
    name="pyved",

    author="moonb3ndr et al.",
    author_email="contact@kata.games",
    url="https://github.com/wkta/gamedev-pyved",

    long_description=long_desc,
    long_description_content_type='text/markdown',

    install_requires=required,

    package_dir={'': 'src'},
    packages=[
        "pyved",
    ],

    package_data={'': ['src/pyved/VERSION']},
    include_package_data=True,  # to be sure we also install non-py files...

    description='experimental toolbox for game devs who use python',
    license='LGPL-3.0',
    version=str(get_version()),

    entry_points={
        'console_scripts': [
            'pyved = pyved.__main__:main'
        ]
    }
)



from typing import Union, List
from pathlib import Path

import pygame

from pygame_gui.windows import UIFileDialog, UIMessageWindow
from pygame_gui import UI_BUTTON_START_PRESS, UI_WINDOW_MOVED_TO_FRONT, UI_WINDOW_CLOSE
from pygame_gui import UI_FILE_DIALOG_PATH_PICKED

from . import utils



# from ui.ui_canvas_window import CanvasWindow
# from ui.ui_new_canvas_dialog import UINewCanvasDialog
# from tools.undo_record import UndoRecord

ABOUT_MSG = '<b>Pyved for Game Development</b><br>'\
            '********<br>' \
            '<b>Authors: </b>moonb3ndr et al.<br>' \
            f"<b>Version: </b>{utils.get_version()}<br>"\
            '<b>License: </b>Gnu LGPL v3.0<br>'


class MenuBarEventHandler:
    def __init__(self, window_surface, ui_manger):

        self.window_surface = window_surface
        self.ui_manager = ui_manger

        self.last_used_file_path = str(Path('.').absolute())

        self.active_canvas_window = None  # type: Union[CanvasWindow, None]

    def process_event(self, event):
        if (event.type == UI_WINDOW_MOVED_TO_FRONT
                and event.ui_object_id == '#canvas_window'):
            self.active_canvas_window = event.ui_element
        if (event.type == UI_WINDOW_CLOSE
                and event.ui_object_id == '#canvas_window'
                and self.active_canvas_window == event.ui_element):
            self.active_canvas_window = None

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#file_menu_items.#new'):
            new_canvas_dialog = pygame.Rect(0, 0, 400, 300)
            new_canvas_dialog.center = self.window_surface.get_rect().center
            UINewCanvasDialog(rect=new_canvas_dialog,
                              manager=self.ui_manager)

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#file_menu_items.#open'):
            file_dialog_rect = pygame.Rect(0, 0, 400, 500)
            file_dialog_rect.center = self.window_surface.get_rect().center
            UIFileDialog(rect=file_dialog_rect,
                         manager=self.ui_manager,
                         window_title="Open...",
                         initial_file_path=self.last_used_file_path,
                         object_id='#open_file_dialog',
                         allow_existing_files_only=True)

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#file_menu_items.#save'
                and self.active_canvas_window is not None
                and self.active_canvas_window.canvas_ui.save_file_path is not None):

            try:
                print("Saving to: " + str(self.active_canvas_window.canvas_ui.save_file_path))
                pygame.image.save(self.active_canvas_window.canvas_ui.get_image(),
                                  str(self.active_canvas_window.canvas_ui.save_file_path))
            except pygame.error:
                message_rect = pygame.Rect(0, 0, 250, 160)
                message_rect.center = self.window_surface.get_rect().center
                message_window = UIMessageWindow(rect=message_rect,
                                                 html_message='Unable to save image.',
                                                 manager=self.ui_manager,
                                                 window_title='Saving error')
                message_window.set_blocking(True)

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#file_menu_items.#save_as'
                and self.active_canvas_window is not None):
            file_dialog_rect = pygame.Rect(0, 0, 400, 500)
            file_dialog_rect.center = self.window_surface.get_rect().center
            if self.active_canvas_window.canvas_ui.save_file_path is not None:
                file_path = self.active_canvas_window.canvas_ui.save_file_path
            else:
                file_path = (Path(self.last_used_file_path) /
                             self.active_canvas_window.window_display_title)
            save_dialog = UIFileDialog(rect=file_dialog_rect,
                                       manager=self.ui_manager,
                                       window_title="Save As...",
                                       initial_file_path=str(file_path),
                                       object_id='#save_file_dialog')
            save_dialog.set_blocking(True)

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#edit_menu_items.#undo'
                and self.active_canvas_window is not None):
            self._try_undo()

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#edit_menu_items.#redo'
                and self.active_canvas_window is not None):
            self._try_redo()

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#view_menu_items.#info'
                and self.active_canvas_window is not None):
            info_window_rect = pygame.Rect(0, 0, 400, 250)
            info_window_rect.center = self.window_surface.get_rect().center

            file_name = self.active_canvas_window.window_display_title
            pixel_size = (str(self.active_canvas_window.canvas_ui.rect.width) +
                          ' x ' +
                          str(self.active_canvas_window.canvas_ui.rect.height) +
                          ' pixels.')

            UIMessageWindow(rect=info_window_rect,
                            html_message='<br><b>Image Info</b><br>'
                                         '---------------<br><br>'
                                         '<b>File Name: </b>' + file_name + '<br>'
                                         '<b>Pixel size: ' + pixel_size + '<br>',
                            manager=self.ui_manager,
                            window_title='Image info')

        if (event.type == UI_BUTTON_START_PRESS
                and event.ui_object_id == 'menu_bar.#help_menu_items.#about'):
            about_window_rect = pygame.Rect(0, 0, 400, 250)
            about_window_rect.center = self.window_surface.get_rect().center
            UIMessageWindow(rect=about_window_rect,
                            html_message=ABOUT_MSG,
                            manager=self.ui_manager,
                            window_title='About')

        if (event.type == UI_FILE_DIALOG_PATH_PICKED and
                event.ui_object_id == '#save_file_dialog'):
            path = Path(event.text)
            self.last_used_file_path = path.parent
            try:
                print("Saving to: " + str(path))
                pygame.image.save(self.active_canvas_window.canvas_ui.get_image(),
                                  str(path))
                self.active_canvas_window.set_display_title(path.name)
                self.active_canvas_window.canvas_ui.save_file_path = path
            except pygame.error:
                message_rect = pygame.Rect(0, 0, 250, 160)
                message_rect.center = self.window_surface.get_rect().center
                message_window = UIMessageWindow(rect=message_rect,
                                                 html_message='Unable to save image to path: '
                                                              '<br>' + str(path) + '<br><br>'
                                                              'pygame can only save to .bmp, .png,'
                                                              '.jpg & .tga. TGA is the default '
                                                              'format.',
                                                 manager=self.ui_manager,
                                                 window_title='Saving error')
                message_window.set_blocking(True)

        if (event.type == pygame.KEYDOWN and event.key == pygame.K_z
                and event.mod & pygame.KMOD_CTRL):
            if event.mod & pygame.KMOD_SHIFT:
                self._try_redo()
            else:
                self._try_undo()

    def _try_undo(self):
        if (self.active_canvas_window is not None
                and self.active_canvas_window.canvas_ui.undo_stack):
            undo_record = self.active_canvas_window.canvas_ui.undo_stack.pop()

            redo_surf = pygame.Surface(undo_record.rect.size,
                                       flags=pygame.SRCALPHA)
            redo_surf.blit(self.active_canvas_window.canvas_ui.get_image(),
                           (0, 0), undo_record.rect)
            redo_record = UndoRecord(redo_surf, undo_record.rect.copy())
            self.active_canvas_window.canvas_ui.redo_stack.append(redo_record)
            self.active_canvas_window.canvas_ui.get_image().blit(undo_record.image,
                                                                 undo_record.rect)

    def _try_redo(self):
        if (self.active_canvas_window is not None
                and self.active_canvas_window.canvas_ui.redo_stack):
            redo_record = self.active_canvas_window.canvas_ui.redo_stack.pop()
            undo_surf = pygame.Surface(redo_record.rect.size,
                                       flags=pygame.SRCALPHA)
            undo_surf.blit(self.active_canvas_window.canvas_ui.get_image(),
                           (0, 0), redo_record.rect)
            undo_record = UndoRecord(undo_surf, redo_record.rect.copy())
            self.active_canvas_window.canvas_ui.undo_stack.append(undo_record)

            self.active_canvas_window.canvas_ui.get_image().blit(redo_record.image,
                                                                 redo_record.rect)
